
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <signal.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>

#include <termios.h>
#include <fcntl.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <netinet/in.h>

#include <bluetooth/bluetooth.h>
#include <bluetooth/hci.h>
#include <bluetooth/hci_lib.h>
#include <bluetooth/sdp.h>
#include <bluetooth/sdp_lib.h>

#include "localserver.h"


//#define DEBUG_BLUESERVER

#ifdef DEBUG_BLUESERVER
  #define LS_DEBUG(msg...)     do {          printf(msg); fflush(stdout); }   while (0)
  #define LS_DEBUGT(t, msg...) do { if (t) { printf(msg); fflush(stdout); } } while (0)
#else
  #define LS_DEBUG(msg...)     do { } while (0)
  #define LS_DEBUGT(t, msg...) do { } while (0)
#endif


// major class of device codes
// taken from: https://www.bluetooth.org/foundry/assignnumb/document/baseband
#define MAJ_COD_MISC                  0
#define MAJ_COD_COMPUTER              1
#define MAJ_COD_PHONE                 2
#define MAJ_COD_LAN                   3
#define MAJ_COD_AUDIO_VIDEO           4
#define MAJ_COD_COMPUTER_PERIPHERAL   5
#define MAJ_COD_IMAGING               6
#define MAJ_COD_UNCATEGORIZED         31

// minor class of device types (where MajorDeviceType is Computer or PDA)
#define MAJ_COMP_MIN_HANDHELD         0x04  // Handheld PC/PDA (clam shell)
#define MAJ_COMP_MIN_PALMSIZED        0x05  // Palm sized PC/PDA

#define DEBUG_CLASS_OF_DEVICE_FILTERING

char outbuf[1024];

int scan_profiles_fd = -1;
typedef int (*handler_t)(bdaddr_t *bdaddr, struct search_context *arg);
static bdaddr_t interface_bdaddr;
int estr2ba(char *str, bdaddr_t *ba)
{
	// Only trap "local", "any" is already dealt with 
	if(!strcmp(str, "local")) 
  {
		bacpy(ba, BDADDR_LOCAL);
		return 0;
	}
	return str2ba(str, ba);
}

// Pass args to the inquiry/search handler 
struct search_context 
{
	char 		*svc;		// Service 
	uuid_t		group;		// Browse group 
	int		tree;		// Display full attribute tree 
	uint32_t	handle;		// Service record handle 
};



static char UUID_str[MAX_LEN_UUID_STR];


static void scan_profiles_print_service_class(void *value, void *userData)
{
	char ServiceClassUUID_str[MAX_LEN_SERVICECLASS_UUID_STR];
	uuid_t *uuid = (uuid_t *)value;

	sdp_uuid2strn(uuid, UUID_str, MAX_LEN_UUID_STR);
	sdp_svclass_uuid2strn(uuid, ServiceClassUUID_str, MAX_LEN_SERVICECLASS_UUID_STR);
  char dummy[128];
	sprintf(dummy,"0x%s|%s|", UUID_str, ServiceClassUUID_str);
  strcat(outbuf,dummy);
}


static void scan_profiles_print_service_desc(void *value, void *user)
{
	char str[MAX_LEN_PROTOCOL_UUID_STR];
	sdp_data_t *p = (sdp_data_t *)value, *s;
	int i = 0, proto = 0;
	
	for (; p; p = p->next, i++) 
  {
		switch (p->dtd) 
    {
		case SDP_UUID16:
		case SDP_UUID32:
		case SDP_UUID128:
			sdp_uuid2strn(&p->val.uuid, UUID_str, MAX_LEN_UUID_STR);
			sdp_proto_uuid2strn(&p->val.uuid, str, sizeof(str));
			proto = sdp_uuid_to_proto(&p->val.uuid);
			break;
    case SDP_UINT8:
			if (proto == RFCOMM_UUID)
      {
        sprintf(outbuf,"Channel|%d|\n", p->val.uint8);
        write(scan_profiles_fd,outbuf,strlen(outbuf));
      }
      break;
    default:
			break;
		}
	}
}


void scan_profiles_print_access_protos(void* vr, void* userData)
{
	sdp_list_t *value = (sdp_list_t *)vr;
	sdp_list_t *protDescSeq = (sdp_list_t *)value;
	sdp_list_foreach(protDescSeq, scan_profiles_print_service_desc, 0);
}


//
// Parse a SDP record in user friendly form.
//
void scan_profiles_print_service_attr(sdp_record_t *rec)
{
	sdp_list_t *list = 0, *proto = 0;

	sdp_data_t *d = sdp_data_get(rec, SDP_ATTR_SVCNAME_PRIMARY);
	if (d) 
  {
    sprintf(outbuf,"Name|%s|\n", d->val.str);
    write(scan_profiles_fd,outbuf,strlen(outbuf));
  }
  else
  {
	  d = sdp_data_get(rec, SDP_ATTR_SVCDESC_PRIMARY);
	  if (d) 
    {
      sprintf(outbuf,"Name|%s|\n", d->val.str);
      write(scan_profiles_fd,outbuf,strlen(outbuf));
    }
    else
    {
  	  d = sdp_data_get(rec, SDP_ATTR_PROVNAME_PRIMARY);
	    if (d) 
      {
        sprintf(outbuf,"Name|%s|\n", d->val.str);   
        write(scan_profiles_fd,outbuf,strlen(outbuf));
      }
    }
  }

//	printf("Service RecHandle: 0x%x\n", rec->handle);

	if (sdp_get_service_classes(rec, &list) == 0) 
  {
//		printf("Service Class ID List:\n");
    sprintf(outbuf,"ServiceClass|");
		sdp_list_foreach(list, scan_profiles_print_service_class, 0);
    strcat(outbuf,"\n");
    write(scan_profiles_fd,outbuf,strlen(outbuf));
		sdp_list_free(list, free);
	}
	if (sdp_get_access_protos(rec, &proto) == 0) 
  {
//		printf("Protocol Descriptor List:\n");
		sdp_list_foreach(proto, scan_profiles_print_access_protos, 0);
		sdp_list_free(proto, (sdp_free_func_t)sdp_data_free);
	}
}


//
// Perform an inquiry and search/browse all peer found.
//
static void scan_profiles_inquiry(handler_t handler, void *arg)
{
	inquiry_info ii[20];
	uint8_t count = 0;
	int i;

	LS_DEBUG("Inquiring ...\n");
	if (sdp_general_inquiry(ii, 20, 8, &count) < 0) 
  {
		LS_DEBUG("Inquiry failed\n");
		return;
	}
	for (i=0; i<count; i++)
		handler((bdaddr_t *)&ii[i].bdaddr, (struct search_context *)arg);
}

/*
 * Search for a specific SDP service
 */
int scan_profiles_do_search(bdaddr_t *bdaddr, struct search_context *context)
{
	sdp_list_t *attrid, *search, *seq, *next;
	uint32_t range = 0x0000ffff;
	char str[20];
	sdp_session_t *sess;

	if (!bdaddr) 
  {
		scan_profiles_inquiry(scan_profiles_do_search, context);
		return 0;
	}
	sess = sdp_connect(&interface_bdaddr, bdaddr, SDP_RETRY_IF_BUSY);
	ba2str(bdaddr, str);
	if (!sess) {
		LS_DEBUG("Failed to connect to SDP server on %s: %s\n", str, strerror(errno));
		return -1;
	}

	attrid = sdp_list_append(0, &range);
	search = sdp_list_append(0, &context->group);
	if (sdp_service_search_attr_req(sess, search, SDP_ATTR_REQ_RANGE, attrid, &seq)) {
		LS_DEBUG("Service Search failed: %s\n", strerror(errno));
		sdp_close(sess);
		return -1;
	}
	sdp_list_free(attrid, 0);
	sdp_list_free(search, 0);

	for (; seq; seq = next) 
  {
		sdp_record_t *rec = (sdp_record_t *) seq->data;
		struct search_context sub_context;

		if (context->tree) 
    {
			/* Display full tree */
//			sdp_printf_service_attr(rec);
		} 
    else 
    {
			/* Display user friendly form */
			scan_profiles_print_service_attr(rec);
		}
//		printf("\n");
		
		if (sdp_get_group_id(rec, &sub_context.group) != -1) 
    {
			/* Set the subcontext for browsing the sub tree */
			memcpy(&sub_context, context, sizeof(struct search_context));
			/* Browse the next level down if not done */
			if (sub_context.group.value.uuid16 != context->group.value.uuid16)
				scan_profiles_do_search(bdaddr, &sub_context);
		}
		next = seq->next;
		free(seq);
		sdp_record_free(rec);
	}
	sdp_close(sess);
//  printf("END\n");
	return 0;
}

/*
 * Browse the full SDP database (i.e. list all services starting from the
 * root/top-level).
 */
int scan_profiles_cmd_browse(char* address)
{
	bacpy(&interface_bdaddr, BDADDR_ANY);

	struct search_context context;
	int opt;

	/* Initialise context */
	memset(&context, '\0', sizeof(struct search_context));
#if 0
	/* We want to browse the top-level/root */
	sdp_uuid16_create(&(context.group), PUBLIC_BROWSE_GROUP);
#else
	/* Some phones don't return any services if you ask them for the public
	 * browse group.  It turns out that the Windows XP Bluetooth browser asks
	 * for L2CAP_UUID, which seems to work for all phones. */
	sdp_uuid16_create(&(context.group), L2CAP_UUID);
#endif

	bdaddr_t bdaddr;
	estr2ba(address, &bdaddr);
	return scan_profiles_do_search(&bdaddr, &context);
}

#define CSR_VARID_ENABLE_TX 0x4007
#define CSR_VARID_DISABLE_TX  0x4008

char *bccmd_err_names[] = {
  "ok",
  "no such variable id",
  "value too big",
  "no value",
  "bad request",
  "no access",
  "read only",
  "write only",
  "error",
  "permission denied",
};

int csr_write_varid_uint16(int dd, uint16_t seqnum, uint16_t varid, uint16_t value)
{
  unsigned char cmd[] = {
    0x02, 0x00,
    0x09, 0x00,
    seqnum & 0xff, seqnum >> 8,
    varid & 0xff, varid >> 8,
    0x00, 0x00,      // status
    0x00, 0x00,      // payload
    0x00, 0x00,
    0x00, 0x00,
    0x00, 0x00
  };

  unsigned char cp[254], rp[254];
  struct hci_request rq;
  uint16_t status;

  memset(&cp, 0x00, sizeof(cp));
  cp[0] = 0xc2;
  memcpy(cp + 1, cmd, sizeof(cmd));

  memset(&rq, 0, sizeof(rq));
  rq.ogf    = OGF_VENDOR_CMD;
  rq.ocf    = 0x00;
  rq.event  = EVT_VENDOR;
  rq.cparam = cp;
  rq.clen   = sizeof(cmd) + 1;
  rq.rparam = rp;
  rq.rlen   = sizeof(rp);

  if(hci_send_req(dd, &rq, 2000) < 0) 
  {
    LS_DEBUG("hci_send_req %d %s\n", errno, strerror(errno));
    return -1;
  }

  if(rp[0] != 0xc2) 
  {
    LS_DEBUG("hci_send_req ret %d %s\n", errno, strerror(errno));
    errno = EIO;
    return -1;
  }

  status = rp[9] + (rp[10] << 8);
  if(status != 0) 
  {
    LS_DEBUG("Illegal request: %s\n", bccmd_err_names[status]);
    return -1;
  }

  return 0;
}

int csr_enable_tx(int dd, uint16_t seqnum)
{
  return csr_write_varid_uint16(dd, seqnum, CSR_VARID_ENABLE_TX, 0);
}

int csr_disable_tx(int dd, uint16_t seqnum)
{
  return csr_write_varid_uint16(dd, seqnum, CSR_VARID_DISABLE_TX, 0);
}

void EnableBluetooth(int aEnable)
{
  LS_DEBUG("EnableBluetooth called with %d\n",aEnable);
  int dev_id = -1;
  int ret=0;
  static int seq=1;
  int dd;

  dev_id = hci_devid("hci0");
  if(dev_id < 0) 
  {
    LS_DEBUG("Invalid device\n");
    return;
  }
  dd = hci_open_dev(dev_id);
  if(dd < 0) 
  {
    LS_DEBUG("Error opening hci device\n");
    return;
  }

  if(aEnable) 
  {
    LS_DEBUG("disabling flight mode - enabling TX\n");
    ret = csr_enable_tx(dd, seq++);
  }
  else
  {
    LS_DEBUG("enabling flight mode - disabling TX\n");
    ret = csr_disable_tx(dd, seq++);
  }

  if(ret)
  {
    LS_DEBUG("ret=%d\n", ret);
  }
  hci_close_dev(dd);
}


static void requesthandler(char* request,int fd)
{
  if (!strncmp(request,"DISCOVER",8))
  {
	  inquiry_info *info = NULL;
	  int num_rsp, length, flags;
	  char addr[18];
	  char name[248];
	  int i, opt, dd;

	  length  = 8;  /* ~10 seconds */
	  num_rsp = 100;
	  flags = IREQ_CACHE_FLUSH;
    int dev_id = -1;
	  if (dev_id < 0) 
    {
		  dev_id = hci_get_route(NULL);
		  if (dev_id < 0) 
      {
			  perror("Device is not available");
        write(fd, "END", strlen("END"));
        return;
		  }
	  }

	  LS_DEBUG("Scanning ...\n");
    int check_count = 0;
REDO:
    num_rsp = hci_inquiry(dev_id, length, num_rsp, NULL, &info, flags);
    if (num_rsp == -1 && check_count < 10)
    {
      sleep(1);
      check_count++;
      goto REDO;
    }

    LS_DEBUG("num_rsp = %d\n",num_rsp);

    if (num_rsp >= 0) 
    {
   	  dd = hci_open_dev(dev_id);
	    if (dd >= 0) 
      {
	      for (i = 0; i < num_rsp; i++) 
        {
          int majorClassOfDevice =  (((info+i)->dev_class[1])&0x1f);
          int minorClassOfDevice = ((((info+i)->dev_class[0])>>2)&0x1f);

          #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
          LS_DEBUG("majorClassOfDevice=%d: ", majorClassOfDevice);
          switch (majorClassOfDevice)
          {
            case MAJ_COD_MISC:                LS_DEBUG("Miscellaneous\n"); break;
            case MAJ_COD_COMPUTER:            LS_DEBUG("Computer (desktop,notebook, PDA, organizers)\n"); break;
            case MAJ_COD_PHONE:               LS_DEBUG("Phone (cellular, cordless, payphone, modem)\n"); break;
            case MAJ_COD_LAN:                 LS_DEBUG("LAN/Network Access point\n"); break;
            case MAJ_COD_AUDIO_VIDEO:         LS_DEBUG("Audio/Video (headset,speaker,stereo, video display, vcr)\n"); break;
            case MAJ_COD_COMPUTER_PERIPHERAL: LS_DEBUG("Peripheral (mouse, joystick, keyboards)\n"); break;
            case MAJ_COD_IMAGING:             LS_DEBUG("Imaging (printing, scanner, camera, display)\n"); break;
            case MAJ_COD_UNCATEGORIZED:       LS_DEBUG("Uncategorized, specific device code not specified\n"); break;
            default:                          LS_DEBUG("Unknown class of device\n"); break;
          }
          #endif
          
          int skipType = 1;
          if (majorClassOfDevice == MAJ_COD_PHONE)
            skipType = 0;
          if (majorClassOfDevice == MAJ_COD_COMPUTER && 
             (minorClassOfDevice == MAJ_COMP_MIN_HANDHELD || 
              minorClassOfDevice == MAJ_COMP_MIN_PALMSIZED)) 
            skipType = 0;

          if (skipType)
          {  
            #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
            LS_DEBUG("Skipping (not a phone or PDA)\n");
            #endif

            continue;
          }
          /*$$$ 
          if (	((((info+i)->dev_class[1])>>8)&0x1f) != 2 )
          {
            continue;
          }
          */

          // If the client sends *any* data, just stop scanning
          {
            int nread;
            ioctl(fd, FIONREAD, &nread);
            if (nread>0)
            {
              LS_DEBUG("Stopping scanning for bluetooth devices\n");
              break;
            }
          }
          
          
          memset(name, 0, sizeof(name));
		      if (hci_read_remote_name(dd, &(info+i)->bdaddr, sizeof(name), name, 100000) < 0)
          {
            #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
            LS_DEBUG("Unable to read device name: hci_read_remote_name failed\n");
            #endif

            // Skip devices with no name
//            strcpy(name, "n/a");
            continue;
          }
		      ba2str(&(info+i)->bdaddr, addr);
          char buf[256];
          snprintf(buf,256,"%s|%s", addr, name);
          buf[255] = '\0';
          write(fd, buf, strlen(buf));
          
          #ifdef DEBUG_CLASS_OF_DEVICE_FILTERING
          LS_DEBUG("name = %s\n", buf);
          #endif
	      }
    	  close(dd);
    	  free(info);
      }
    }
    write(fd, "END", strlen("END"));
  }
  else if (!strncmp(request,"PROFILES",8))
  {
    scan_profiles_fd = fd;
    LS_DEBUG("Scanning [%s]\n",&request[9]);
    scan_profiles_cmd_browse(&request[9]);
    write(fd, "END", strlen("END"));
  }
  else if (!strncmp(request,"ENABLE",6))
  {
    EnableBluetooth(1);
    write(fd, "END", strlen("END"));
  }
  else if (!strncmp(request,"DISABLE",7))
  {
    EnableBluetooth(0);
    write(fd, "END", strlen("END"));
  }
  else
  {
    write(fd, "END", strlen("END"));
  }
}

int main(int argc,char** argv)
{
  if (argc<2) exit(-1);
  char* the_address = argv[1];
  return localserver(the_address,requesthandler);
}
